/* ReSCAL - Stochastic simulation engine
 *
 * Copyright (C) 2011
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * Code based on dissol program,
 * by Eduardo Sepulveda <edo@espci.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <memory.h>
#include <math.h>
#include <time.h>

#include "defs.h"
#include "macros.h"
#include "param.h"
#include "format.h"
#include "cells.h"
#include "space.h"
#include "surface.h"
#include "doublets.h"
#include "transitions.h"
#include "simul.h"
#include "simul_models.h"
#include "schedul.h"
#include "trace.h"
#include "lgca.h"
#include "callbacks.h"
#include "interface.h"
#include "view.h"
#ifdef PARALLEL
#include "synchro.h"
#endif

extern unsigned char opt_log;
extern int H, L, D, HL, HLD;       // les dimensions de la terre
extern int *db_pos[];                   // les tableaux contenant la position des doublets actifs
extern int Ncel[];                       // nombre de cellules par type
extern int Ndb[];                        // nombre doublets actifs par type
extern int nb_trans_db, nb_trans_cel, nb_trans; // nombre de transitions de 2 cellules, d'1 cellule et au total
extern TransitionDb t_trans[];            // la description des transitions de doublets
extern TransitionCel t_trans_cel[];           // la description des transitions de cellules
extern LienTransDb   t_lien[];        // la description des liens de correlation entre transitions de 2 cellules
extern int nb_liens;             // nombre de liens entre transitions
#ifdef INFO_TRANS
extern int cpt_lien[];      // compteur des correlations effectives
#endif
extern int direction[];             // conversion classe de transition -> direction

extern int use_lgca;
#ifdef LGCA
extern int flag_lgca_reset;
extern int lgca_iter;    //nombre de cycles de collisions
#endif

#ifdef PARALLEL
extern int mode_par;    //mode parallele
#endif

#ifdef CGV
extern float cgv_coef;
#endif

extern int rot_mode;

#ifdef TIME_SCALE
extern float time_scale; //t_0
#endif

extern int vdir_mode; //display mode of the current orientation
extern unsigned char reorient_flag;

unsigned long int iter=0, md_iter=0;  // nombre d'iterations (unites, milliards d'unites)
double init_time=0.0;                 // initial time
double csp_time=0.0;                  // time variable for the cellular space
double stop_time=0.0;                 // stopping time
#ifdef TIME_SCALE
extern double real_time;       // real time with time scale
#endif
char end_of_simul=0;
int cpt_trans_blocked[MAX_TRANSITIONS_DB];  //nombre de transitions blockees
double PII, PII_time;     //probabilités totales
double PI_ij[MAX_TRANSITIONS];                  // les PI_ij
double cum[MAX_TRANSITIONS];     //probabilites cumulees

char *boundary_str = NULL;
#ifdef CYCLAGE_HOR
int boundary = BC_PERIODIC;  //boundary conditions
int pbc_mode = 1; //periodic boundary conditions
#else
int boundary = BC_REINJECTION;  //boundary conditions
int pbc_mode = 0; //periodic boundary conditions
#endif

unsigned char csphpp_flag = 0;
float dump_delay_png = 0.0;
float dump_delay_csp = 0.0;
double stop_delay_t0 = 0.0;

unsigned char flag_comp_time=1; /// always log computing time

int flag_csp_altered=0; /// cellular space has been altered
#ifdef TIME_CORRECTION
#ifdef CGV
int time_corr=TC_CGV; /// calculate the average shear stress for time correction
#else
int time_corr=TC_MINI; /// no time correction (faster)
#endif
char *time_corr_str=NULL;
char flag_time_corr=0;
float delay_time_corr=10.0;
#endif // TIME_CORRECTION
int flag_save_auto=0; /// save data before each rotation
int flag_dcsp_auto=0; /// save CSP data before each rotation
int flag_dpng_auto=0; /// save images before each rotation

void compute_prob_dist();
int simul_trans();
int simul_check(int , int , int , int);
void simul_lien_trans(int , int , int );
void simul_stop();
void simul_save_auto();

/// Declare some parameters of the automaton
void params_simul()
{
  //csp_time
  param_family("TIME", "Time parameters");
  parameter("Time", "initial time (t0) - optional", &init_time, PARAM_DOUBLE, "TIME");

#ifdef CYCLAGE_HOR
#ifdef MODEL_DUN
  parameter("Boundary", "boundary conditions (PERIODIC|OPEN|OUT|CLOSE|REINJECTION) - PERIODIC by default", &boundary_str, PARAM_STRING, "GENERAL");
#else
  parameter("Boundary", "boundary conditions (PERIODIC|OPEN|CLOSE) - PERIODIC by default", &boundary_str, PARAM_STRING, "GENERAL");
#endif
#elif defined(MODEL_DUN)
  parameter("Boundary", "boundary conditions (OPEN|REINJECTION) - REINJECTION by default", &boundary_str, PARAM_STRING, "GENERAL");
#else
  parameter("Boundary", "boundary conditions (OPEN|CLOSE) - CLOSE by default", &boundary_str, PARAM_STRING, "GENERAL");
#endif //CYCLAGE_HOR

#ifdef TIME_CORRECTION
#ifdef CGV
  parameter("Time_corr", "reliability of time correction (ACCURATE|AVERAGE|MINIMAL) - AVERAGE by default", &time_corr_str, PARAM_STRING, "TIME");
#else
  parameter("Time_corr", "reliability of time correction (ACCURATE|MINIMAL) - MINIMAL by default", &time_corr_str, PARAM_STRING, "TIME");
#endif // CGV
  parameter("Time_corr_delay", "recalculation delay of the effective transition rates for time correction (t0)", &delay_time_corr, PARAM_FLOAT, "TIME");
#endif // TIME_CORRECTION

  params_simul_models();
}

/// Parse parameter values for the automaton
void simul_parse()
{
  if (boundary_str){
    if (!strcmp(boundary_str, "PERIODIC")){
#ifdef CYCLAGE_HOR
      boundary = BC_PERIODIC;
#else
      ErrPrintf("Incorrect value for Boundary: %s, please define CYCLAGE_HOR in defs.h and recompile\n", boundary_str);
      exit(-2);
#endif
    }
    else if (!strcmp(boundary_str, "OPEN")){
      boundary = BC_OPEN;
    }
    else if (!strcmp(boundary_str, "OUT")){
      boundary = BC_OUT;
    }
    else if (!strcmp(boundary_str, "CLOSE")){
      boundary = BC_CLOSE;
    }
    else if (!strcmp(boundary_str, "REINJECTION")){
#ifdef MODEL_DUN
      boundary = BC_REINJECTION;
#else
      ErrPrintf("Incorrect value for Boundary: %s, not compatible with model %s\n", boundary_str, MOD_NAME);
      exit(-2);
#endif
    }
    else{
      ErrPrintf("Incorrect value for Boundary: %s\n", boundary_str);
      exit(-2);
    }
    pbc_mode = (boundary == BC_PERIODIC);
#ifdef PARALLEL
    if (mode_par) pbc_mode = 0;
#endif
  }


#ifdef TIME_CORRECTION
  if (time_corr_str){
    if (!strcmp(time_corr_str, "ACCURATE")){
      time_corr = TC_MAXI;
    }
#ifdef CGV
    else if (!strcmp(time_corr_str, "AVERAGE")){
      time_corr = TC_CGV;
    }
#endif
    else if (!strcmp(time_corr_str, "MINIMAL")){
      time_corr = TC_MINI;
    }
    else{
      ErrPrintf("Incorrect value for Time_corr: %s\n", time_corr_str);
      exit(-2);
    }
  }
  if (!param_is_set("Time_corr_delay")){
    if (time_corr == TC_MAXI){
      delay_time_corr = 10.0;
    }
    else if (time_corr == TC_CGV){
      delay_time_corr = 1.0;
    }
  }
#endif // TIME_CORRECTION

  parse_simul_models();
}

void init_simul()
{
  //params_simul();
  //params_modele();
  simul_parse();
  schedul_init();
}


/// Computate the probability distribution of all transitions
inline void compute_prob_dist()
{
  int i;
  double PI_ij_time, ccc;
  TransitionDb *ptr;
  TransitionCel *ptr_cel;

  //LogPrintf("calcul_prob: start\n");

  // doublet transitions
  ptr = t_trans;
  for(i=0; i<nb_trans_db; i++, ptr++){
    //regulation of a transition
    if (ptr->regul) ptr->regul(&ptr->intensite);
    PI_ij[i] = (double) Ndb[ptr->depart]*ptr->intensite;
#ifdef CGV
    //if (t_trans[i].check) PI_ij[i] *= cgv_coef;
#endif
    //if (PI_ij[i]) LogPrintf("tr = %d   rate = %f   time = %f\n", i, ptr->intensite, csp_time); //debug
  }

  // cell transitions
  ptr_cel = t_trans_cel;
  for(; i<nb_trans; i++,ptr_cel++){
    if (ptr_cel->regul) ptr_cel->regul(&ptr_cel->intensite); //regulation de la transition
    PI_ij[i] = (double) Ncel[ptr_cel->depart] * ptr_cel->intensite;
  }

  // calcul PII : overall probability for any transition to occur in a unit of time
  // calcul PII_time : the same only for the transitions with time evolution (and time correction if transitions are blocked)
  for(i=0, PII=0.0, ccc=0.0, PII_time = 0.0; i<nb_trans; i++){
    PII += PI_ij[i];
    if (i<nb_trans_db){
      if (t_trans[i].time_mode == TIME_EVOL)
        PI_ij_time = PI_ij[i];
#ifdef TIME_CORRECTION
      else if (t_trans[i].time_mode == TIME_CORR){
        if (time_corr == TC_MAXI)
          PI_ij_time = PI_ij[i]*t_trans[i].chk_prob;
  #ifdef CGV
        else if (time_corr == TC_CGV)
          PI_ij_time = PI_ij[i]*cgv_coef;
  #endif // CGV
        else
          PI_ij_time = PI_ij[i];
      }
#endif // TIME_CORRECTION
      else
        PI_ij_time = 0;
    }
    else{
      PI_ij_time = PI_ij[i];
    }
/*#ifdef CGV
    //if ((i<nb_trans_db) && (t_trans[i].check != NULL)) PII_time  += PI_ij[i]*cgv_coef; else PII_time  += PI_ij[i];
    if ((i<nb_trans_db) && (t_trans[i].time_mode == TIME_CORR)) PII_time  += PI_ij[i]*cgv_coef; else PII_time  += PI_ij[i];
#endif*/
    cum[i] = PI_ij[i] + ccc;
    ccc    = cum[i];
    PII_time += PI_ij_time;
  }
/*#ifndef CGV
  PII_time = PII;
#endif*/

  //LogPrintf("PII = %f\n", PII);
  //LogPrintf("PII_time = %f\n", PII_time);
  //return PII;
}

/// Perform time evolution and check for events
void set_time(double t)
{
  double dt, th;

  //LogPrintf("set time to %f\n", t);

  if (get_total_nb_events() > 0){
    /// threshold of next event
    th = get_next_time_threshold();

    while (th<=t){
      //LogPrintf("set_time %f -> %f, th = %f\n", csp_time, t, th);
#ifdef TIME_SCALE
      if (use_lgca) real_time += (th-csp_time)*time_scale;
#endif
      /// trigger next event
      csp_time = th;
      schedule_family(EF_ALL);
      th = get_next_time_threshold();
    }
    if (flag_csp_altered) return;
  }

#ifdef TIME_SCALE
  if (use_lgca) real_time += (t-csp_time)*time_scale;
#endif

  csp_time = t;
}

void simul_time()
{
  double aleat, dt;

  //computation of csp and real time evolution
  //using Poisson formula
  if (PII_time){
    aleat = drand48();
    if (!aleat) {ErrPrintf("ERROR: simul_time - aleat = 0 !\n"); exit(-1); }
    dt = -log(aleat)/PII_time;
    set_time(csp_time + dt);
  }
  else{
    LogPrintf("simul_time: PII_time = 0, csp_time = %g\n", csp_time);
  }
}

/// Estimate the effective rate of each transition for accurate time correction
#ifdef TIME_CORRECTION
void init_time_corr()
{
  if (time_corr != TC_MINI){
    flag_time_corr = trans_time_corr();
    schedul_add_reg_event(EVT_TIME_CORR, delay_time_corr, 0);

    if (time_corr == TC_MAXI){
      LogPrintf("accurate time correction\n");
    }
    else{
      LogPrintf("rough estimation of time correction using shear stress\n");
    }
    LogPrintf("flag_time_corr = %d\n", flag_time_corr);
    LogPrintf("delay_time_corr = %f\n", delay_time_corr);
  }
  else{
    LogPrintf("no time correction\n");
  }
}

void simul_time_corr()
{
  int i, it, ix, dir, nb, db_depart;

//#ifdef OPENMP
//#pragma omp parallel for private(it,db_depart,i,ix,dir) reduction(+:nb)
//#endif
  for(it=0; it<nb_trans_db; it++){
    //if (t_trans[it].nb_chk){
    if (t_trans[it].time_mode == TIME_CORR){
      db_depart = t_trans[it].depart;
      /// loop on each doublet that matches the initial doublet of the transition
      nb = 0;
      for(i=0; i<Ndb[db_depart]; i++){
        /// position
        ix = db_pos[db_depart][i];
        /// direction
        if (t_trans[it].classe == HORIZONTAL)
          dir = dir_hor_doublet(ix, db_depart, i);
        else
          dir = direction[t_trans[it].classe];

        if (simul_check(it,ix,dir,0)) nb++;
      }
      t_trans[it].chk_prob = (Ndb[db_depart]>0) ? (float)nb/Ndb[db_depart] : 1;
    }
  }
}
#endif // TIME_CORRECTION

/// Select the next transition to occur and compute time evolution accordingly
inline int simul_trans()
{
  int i, elt, tr;
  double ccc, aleat;
  int dir;
  static int cpt_blk=0;
  int max_blk=HLD;
  char flag_evol_time=0;
  char flag_trans_done=0;
  char flag_trans_canceled=0;
  char time_blocked=0;

  flag_csp_altered = 0;

  while ((!flag_trans_done) && (!flag_trans_canceled) && (cpt_blk < max_blk))
  {
    /// we choose randomly which transition will occur
    tr = 0;
    aleat = drand48();
    ccc = PII * aleat;
    while (ccc >= cum[tr]) tr++;

    //LogPrintf("tr=%d\n",tr);//debug

    if (tr > nb_trans){
      ErrPrintf("ERROR: array overflow in t_trans (%d)\n", tr);
      exit(-1);
    }

    if (tr < nb_trans_db){
      /// DOUBLET TRANSITION

      /// we choose randomly which doublet will undergo the transition
      int db_depart = t_trans[tr].depart;
      elt = (int) floor( drand48() * Ndb[db_depart]);

      int ix = db_pos[db_depart][elt];

  /*#ifdef TRACE
      if (ii>10) trace_test(ix);
  #endif*/

      /// direction (class) of the transition
      if (t_trans[tr].classe == HORIZONTAL)
        dir = dir_hor_doublet(ix, db_depart, elt);
      else
        dir = direction[t_trans[tr].classe];

      /// control of the transition
      if (t_trans[tr].nb_chk && !simul_check(tr, ix, dir, 1)){
        /// blocked transition
        //LogPrintf("transition blocked!\n");
        cpt_blk++;
        continue;
      }

      /// time evolution
      flag_evol_time = (t_trans[tr].time_mode != TIME_NO_EVOL);
      if (flag_evol_time) simul_time();
      //LogPrintf("csp_time = %f\n", csp_time);

      if (flag_csp_altered){
        //LogPrintf("transition canceled!\n");
        flag_trans_canceled = 1;
      }
      else{
        do_trans_db(tr, ix, dir);
        flag_trans_done = 1;
        iter++; //LogPrintf("iter = %ld\n", iter);
        if (iter > 1000000000L){
          iter -= 1000000000L; /// support for 32 bits CPU
          md_iter++;
        }
#ifdef LIENS_TRANSITIONS
        /// linked transitions
        if (t_trans[tr].lien) simul_lien_trans(tr, ix, dir);
#endif
      }

      if (flag_evol_time) cpt_blk = 0; /// reset counter when last transition makes time evolve
    }

    else{
      /// CELL TRANSITION
      static int imax = 0;
      int ix;
      //LogPrintf("tr = %d   aleat = %f   ccc = %f   PII = %f\n", tr, aleat, ccc, PII);

      /// time evolution
      simul_time();

      if (flag_csp_altered){
        flag_trans_canceled = 1;
      }
      else{
        tr -= nb_trans_db;
        i=0;
        /// we look for a cell of the given type
        do{
          ix = (int) floor( drand48() * HLD);
          i++;
        }
        while (get_celltype(ix) != t_trans_cel[tr].depart);

        if (i>imax){
          imax = i;
          if (imax>100){
            LogPrintf ("WARNING: cell transition was hard to achieve (%d trials)\n", imax);
          }
        }

        do_trans_cel(tr, ix);
        iter++; //LogPrintf("iter = %ld\n", iter);
        flag_trans_done = 1;
      }
      cpt_blk = 0;
    }
  }

  //LogPrintf("cpt_blk = %d\n", cpt_blk);

  time_blocked = (cpt_blk >= max_blk);

  if (time_blocked){
    LogPrintf("simul_trans: too many blocked transitions (%d)\n", cpt_blk);
    //log_info(); //debug
    cpt_blk = 0;
  }

  return time_blocked;
}

/// Select and perform a linked transition involving one of the cells
/// of a given doublet and oriented in a given direction
void simul_correl(int tr1, int ix, int ix2, int dir)
{
  static double cum2[MAX_LIENS];
  static int ln[MAX_LIENS], ind[MAX_LIENS];
  double ccc, aleat;
  int i, db, db2, n;
  int trans2_db_depart;
  //determination du doublet de depart
  db = type_doublet(ix, dir);
  db2 = type_doublet(ix2, dir);

  //construction du tableau des transitions correlees
  for(i=0, ccc=0.0, n=0; i<nb_liens; i++){
    if (t_lien[i].trans1 == tr1){
      trans2_db_depart = t_trans[t_lien[i].trans2].depart;
      if (t_lien[i].cel == 1){
        if (trans2_db_depart == db){
          ccc += t_lien[i].intensite;
          cum2[n] = ccc;
          ln[n] = i;
          ind[n] = ix;
          n++;
        }
      }
      else{
        if (trans2_db_depart == db2){
          ccc += t_lien[i].intensite;
          cum2[n] = ccc;
          ln[n] = i;
          ind[n] = ix2;
          n++;
        }
      }
    }
  }
  if (n > 0){ //test tableau non-vide ?
    //choix d'une transition selon les probabilites cumulees
    if (ccc > 1.0){
      ErrPrintf("ERROR : cumulated probabilites = %f > 1.0\n",ccc);
      exit(-1);
    }
    aleat = drand48();
    i=0;
    while ((i < n) && (aleat > cum2[i])) i++;
    if (i < n){
      //LogPrintf("Lien %d -> %d\n", tr1, ltr[i]);
      int tr2 = t_lien[ln[i]].trans2;
      /*if (t_trans[tr2].nb_chk){
        //controle de la transition
        if (!simul_check(tr2, ind[i], dir)) return; //transition bloquee
      }*/
      do_trans_db(tr2, ind[i], dir);
      iter++;
#ifdef INFO_TRANS
      cpt_lien[ln[i]]++;
#endif
      if (t_trans[tr2].lien) simul_lien_trans(tr2, ind[i], dir);
    }
  }
}

/// Perform a chain (or link) of transitions with a given probability
void simul_lien_trans(int tr, int ix, int dir)
{
  int ix2;
  int ix_west, ix2_west;
  int ix_up, ix2_up;
  int ix_north, ix2_north;
  int x,y,z;
  Pos2 *cp;

  if (t_trans[tr].classe == VERTICAL){
    // PREMIERE TRANSITION VERTICALE
    ix2 = get_cell_down(ix);

    // liens horizontaux, doublet est
    simul_correl(tr, ix, ix2, EST);

    // liens horizontaux, doublet ouest
    ix_west = get_cell_west(ix);
    ix2_west = get_cell_down(ix_west);
    simul_correl(tr, ix_west, ix2_west, EST);

    // liens horizontaux, doublet sud
    simul_correl(tr, ix, ix2, SUD);

    // liens horizontaux, doublet nord
    ix_north = get_cell_north(ix);
    ix2_north = get_cell_down(ix_north);
    simul_correl(tr, ix_north, ix2_north, SUD);
  }
  else {
    // PREMIERE TRANSITION HORIZONTALE
    ix2 = get_cell_dir(ix, dir);
    int flag_eo = (dir == EST/*EST_OUEST*/);

    // liens verticaux, doublet bas
    simul_correl(tr, ix, ix2, BAS);

    // liens verticaux, doublet haut
    ix_up = get_cell_up(ix);
    ix2_up = get_cell_up(ix2);
    simul_correl(tr, ix_up, ix2_up, BAS);

    if (!flag_eo){
      // liens horizontaux, doublet est
      simul_correl(tr, ix, ix2, EST);

      // liens horizontaux, doublet ouest
      ix_west = get_cell_west(ix);
      ix2_west = get_cell_west(ix2);
      simul_correl(tr, ix_west, ix2_west, EST);
    }
    else{
      // liens horizontaux, doublet sud
      simul_correl(tr, ix, ix2, SUD);

      // liens horizontaux, doublet nord
      ix_north = get_cell_north(ix);
      ix2_north = get_cell_north(ix2);
      simul_correl(tr, ix_north, ix2_north, SUD);
    }
  }
}

/// Check whether the current transition should be blocked
/// by using its callback function
int simul_check(int tr, int ix, int dir, int flag_cpt)
{
  int i;
  int chk = 1;
  DataCheck *pdc;

  pdc = t_trans[tr].checks;

  for(i=0;  (i<t_trans[tr].nb_chk) && chk; i++, pdc++){
    pdc->dir = dir;

    /// callback function on first cell
    if (pdc->cel == 1){ // test sur la premiere cellule
      chk = pdc->func(ix, (void*)pdc);
    }
    else{ /// callback function on second cell
      int ix2 = get_cell_dir(ix, dir);
      chk = pdc->func(ix2, (void*)pdc);
    }

    /// inversion flag
    if (pdc->inv) chk = !chk;
  }

  if (flag_cpt && !chk){
    cpt_trans_blocked[tr]++;
#ifdef TRACE_TRANS
    trace_trans_blocked(tr, ix);
#endif
  }

  return chk;
}


void simul_stop()
{
  if (stop_time && (csp_time >= stop_time)){
    /// trigger remaining save events
    //schedule_family(EF_GEN);
    //LogPrintf("stop_time=%.10f\n", stop_time);
    LogPrintf("csp_time=%lf\n", csp_time);
    LogPrintf("time to quit\n");
    push_status("stopping ...", 0);
    sleep(1);
    exit(0);
  }
}


int simul_csp()
{
  int ii;
  int time_blocked = 0;

  /// start time
  if (param_is_set("Time")) csp_time = init_time;
  LogPrintf("csp_time = %lf\n", csp_time);

  /// stop time
  if (stop_delay_t0){
    stop_time = csp_time + stop_delay_t0;
  }
#ifdef DUMP_SIGNATURE
  if (opt_log) dump_signature(0);
#endif

  /// initialization of model specific events and data
  init_simul_models();

#ifdef TIME_CORRECTION
  /// initialization of time correction
  init_time_corr();
#endif // TIME_CORRECTION

  if (stop_time){
    schedul_add_event(EVT_STOP, stop_time);
    LogPrintf("stop_time = %f\n", stop_time);
  }

  /// CSP ENGINE START
  /// reset counters
  ii = 0;
  ResetMemory(cpt_trans_blocked, int, MAX_TRANSITIONS_DB);

  lock_csp(0);

  /// trigger "model" events (centering, rotation, lgca ...) before the main loop
  if (get_nb_events(EF_MODEL)) schedule_family(EF_MODEL);

  ///////////////////////// MAIN LOOP /////////////////////////
  LogPrintf("main loop\n");

  while(!end_of_simul)
  {
#ifdef GUI
    /// multithreading
    do_thread_sched();
#endif

    //for (i=0; i<100000/L; i++) compute_alti(GR, ALTI_MODE_BAS); /// slow down ...

#ifdef LGCA
    if (use_lgca && flag_lgca_reset){
      /// reinitialization and stabilization of lattice gas
      simul_lgca_coupling();
    }
#endif // LGCA

    /// STOCHASTIC ENGINE

    /// compute the probability distribution
    compute_prob_dist();

    if (PII == 0){
      /// no active doublets
      LogPrintf("no active doublets (PII=0, csp_time=%g)\n", csp_time);
      time_blocked = 1;
      if (!use_lgca){
        /// stop simulation
        LogPrintf("stopping simulation ...\n");
        end_of_simul = 1;
        break;
      }
    }

    if (PII > 0){
      /// try to perform a stochastic transition
      time_blocked = simul_trans();
#ifdef MODEL_LIFE
      if (time_blocked){
        /// no transition performed
        LogPrintf("stopping simulation ...\n");
        end_of_simul = 1;
        break;
      }
#endif
    }

#ifdef LGCA
    if (time_blocked && (!end_of_simul) && use_lgca){
      /// if time cannot evolve through transitions and lattice gas is enabled, we use the elementary time scale of lattice gas
      double t = get_next_time_threshold();
      //LogPrintf("time evolution forced (%f -> %f)\n", csp_time, t);
      set_time(t);
    }
#endif

    /// specific processes of the model
    simul_model(ii);

    /// counter
    ii++;

#ifdef DUMP_SIGNATURE
    if (opt_log &&  !(ii & 0x000fffff)){ /// every 1048576 transitions
    	/// save fingerprint of the cellular space (hash function)
      dump_signature(ii);
    }
#endif

#ifdef PARALLEL
    //if (mode_par && !(ii & 0xf)) {
    if (mode_par && !(ii & 0xff)) {
      //if (synchro_terre()) {
        //LogPrintf("pause synchro\n");
        while (synchro_terre());
        //LogPrintf("synchro ok\n");
      //}
    }
#endif

  }
  //END OF MAIN LOOP

  unlock_csp(0);

  LogPrintf("out of main loop\n");

#ifdef LISSAGE
  lissage_terre();
#endif

  //return (PII != 0);
  return (end_of_simul);
}


void simul_save_data(int type)
{
  static int save_counter = 0;
  static double t0=0, t1=0;
  static char name[100];
  static char str[50];
  static char start=1;
  static double last_csp_time = 0.0;

  if (start){
#ifdef PARALLEL
    extern int proc_id;
    sprintf(str,"_%d",proc_id);
#else
    *str = 0;
#endif
    start = 0;
  }
  else if (csp_time > last_csp_time){
    //update counter
    save_counter++;
    save_counter = save_counter % 100000; //no more than 100000 files, for security
  }

  elapsed(&t1);

  ///no more than 10 files per sec., for security
  /*if (!t0 || (t1-t0 >= 0.1))*/{
    lock_display(0);

  #ifdef REORIENT_AUTO
    reorient_flag = 1;
    rotation(0, rot_mode, ROT_REORIENT_TEMP);
    //display wind direction
    vdir_mode = VDIR_WIND;
  #endif

    if (type == EVT_SAVE_PNG){
  #if !defined(USE_LIBPNG) && !defined(USE_GD)
      gcallback_flash(0); //update image buffer
  #endif
      sprintf(name,"%s%05d_t0%s.png", MOD_NAME, save_counter, str);
      dump_image(name, "png");
    }

    if (type == EVT_SAVE_CSP){
      dump_terre(DUMP_CSP, save_counter, UNIT_T0);
  #ifdef ALTI
      dump_surface("ALTI", save_counter, UNIT_T0);
  #endif
  #if defined(LGCA) //&& !defined(ROTATIONS)
      if (use_lgca && csphpp_flag){
        dump_mvt(save_counter, UNIT_T0);
  #ifdef CGV
        dump_grad_vel(save_counter, UNIT_T0);
  #endif
      }
  #endif
    }

  #ifdef REORIENT_AUTO
    rotation(0, rot_mode, ROT_REORIENT_UNDO);
    vdir_mode = VDIR_NONE;
    reorient_flag = 0;
  #endif

    unlock_display(0);

    last_csp_time = csp_time;

    t0 = t1;
  }

}

void conv_time_human(double t, char *str)
{
  long i=(long)floor(t);
  int is = i % 60;
  int im = (i/60) % 60;
  int ih = (i/3600) % 24;
  int id = (i/86400);
  sprintf(str, "%03dd%02dh%02dm%02ds", id, ih, im, is);
}


void dump_time()
{
  static int cpt = 0;
  static unsigned long int md_iter_0 = 0, iter_0 = 0;
  static double csp_time_0 = 0;
  static double real_time_0 = 0;
  static double comp_time = 0;
  static double comp_time_0 = 0;
  static double comp_time_start = 0;
  static int lgca_iter_0 = 0;
  long int delta_iter, delta_md_iter;
  FILE *fp;

  fp = fopen("TIME.log","a");
  if (!cpt){
    fprintf(fp,"      nb trans.        delta trans.    time            delta time       ");
#ifdef LGCA
    if (use_lgca) fprintf(fp,"lgca cyc.    ");
#endif
    csp_time_0 = csp_time;
#ifdef TIME_SCALE
    if (use_lgca){
      fprintf(fp,"real time      delta real time   ");
      real_time_0 = real_time;
    }
#endif
    if (flag_comp_time){
      fprintf(fp,"comp. time      delta comp. time   ");
      elapsed(&comp_time_start);
      comp_time = comp_time_0 = comp_time_start;
    }
  }
  delta_iter = iter - iter_0;
  delta_md_iter = md_iter - md_iter_0;
  if ((delta_md_iter > 0) && (delta_iter < 0)) /// for 32 bits CPU
  {
    delta_iter += 1000000000;
    delta_md_iter--;
  }

  if (delta_md_iter)
    fprintf(fp,"\n%04d: %04lu%09lu %03lu%09lu       %e    %e     ", cpt, md_iter, iter, delta_md_iter, delta_iter, csp_time, csp_time - csp_time_0);
  else
    fprintf(fp,"\n%04d: %04lu%09lu    %09lu       %e    %e     ", cpt, md_iter, iter, delta_iter, csp_time, csp_time - csp_time_0);
#ifdef LGCA
  if (use_lgca){
    fprintf(fp,"%09d    ", lgca_iter - lgca_iter_0);
    lgca_iter_0 = lgca_iter;
  }
#endif
#ifdef TIME_SCALE
  if (use_lgca){
    fprintf(fp,"%e   %e      ", real_time, real_time - real_time_0);
  }
#endif
  if (flag_comp_time){
    if (cpt) elapsed(&comp_time);
    char str[50];
    conv_time_human(comp_time - comp_time_start, str);
    fprintf(fp,"%s   %010.02f", str, comp_time - comp_time_0);
    comp_time_0 = comp_time;
  }
  fclose(fp);

  md_iter_0 = md_iter;
  iter_0 = iter;
  csp_time_0 = csp_time;
#ifdef TIME_SCALE
  real_time_0 = real_time;
#endif

#ifdef TIME_CORRECTION
  int i;
  if ((time_corr == TC_MAXI) && flag_time_corr){
    fp = fopen("TIME_CORR.log","a");
    if (!cpt){
      fprintf(fp,"\n# PROBABILITY OF THE TRANSITIONS FOR TIME CORRECTION\n");
      fprintf(fp,"\n     ");
      for (i=0; i<nb_trans_db; i++){
        if (t_trans[i].time_mode == TIME_CORR){
          fprintf(fp,"     TR(%2d)", i);
        }
      }
    }
    else{
      fprintf(fp,"\n%04d:", cpt);
      for (i=0; i<nb_trans_db; i++){
        if (t_trans[i].time_mode == TIME_CORR){
          fprintf(fp,"     %.4f", t_trans[i].chk_prob);
        }
      }
    }
    fclose(fp);
  }
#ifdef CGV
  else if (time_corr == TC_CGV){
    dump_cgv_coef();
  }
#endif // CGV
#endif // TIME_CORRECTION

  cpt++;
}

extern int cpt_trans_db[];       // compteur de transitions effectives de doublet
extern int cpt_trans_cel[];

void dump_trans_rate()
{
  static int cpt = 0;
  FILE *fp;
  int i, db_depart, type ,nb;
  static double csp_time_0 = 0;
  float delay, rate;

  fp = fopen("RATE.log","a");

  if (!cpt){
    fprintf(fp,"\n# ESTIMATION OF EFFECTIVE TRANSITION RATES\n");
    fprintf(fp,"\n    ");
    for (i=0; i<nb_trans_db; i++){
      fprintf(fp,"     TR(%2d)", i);
    }
    for (i=0; i<nb_trans_cel; i++){
      fprintf(fp,"    TR1(%2d)", i);
    }
  }
  else{
    delay = csp_time - csp_time_0;
    fprintf(fp,"\n%04d:",cpt);
    for (i=0; i<nb_trans_db; i++){
      db_depart = t_trans[i].depart;
      nb = Ndb[db_depart];
      rate = nb ? (float)cpt_trans_db[i]/(nb*delay) : 0;
      fprintf(fp,"   %08.3f", rate);
    }
    for (i=0; i<nb_trans_cel; i++){
      type = t_trans_cel[i].depart;
      nb = Ncel[type];
      rate = nb ? (float)cpt_trans_cel[i]/(nb*delay) : 0;
      fprintf(fp,"   %08.3f", rate);
    }
  }

  fclose(fp);

  csp_time_0 = csp_time;
  cpt++;
}


void simul_save_auto()
{
  if (flag_dcsp_auto) simul_save_data(EVT_SAVE_CSP);

  if (flag_dpng_auto) simul_save_data(EVT_SAVE_PNG);
}

void simul_general_event(int type)
{
  //double delay;

  //LogPrintf("event %d triggered\n", p_evt->type);

  /// handle an event of type "save"
  switch (type){
    case EVT_LOG:
      log_info();
      break;

    case EVT_SAVE_CSP:
    case EVT_SAVE_PNG:
      simul_save_data(type);
      break;

#ifdef TIME_CORRECTION
    case EVT_TIME_CORR:
      if (time_corr == TC_MAXI)
        simul_time_corr();
#ifdef CGV
      else if (time_corr == TC_CGV)
        compute_coef_cgv();
#endif // CGV
      break;
#endif // TIME_CORRECTION

    case EVT_STOP:
      simul_stop();
      break;

    default:
      ErrPrintf("ERROR: event %d not supported in family SAVE\n", type);
      exit(-1);
  }
}
